/* 
 *   Creation Date: <2000/05/28 02:37:51 samuel>
 *   Time-stamp: <2003/08/19 12:56:57 samuel>
 *   
 *	<osi_sound.c>
 *	
 *	Sound Driver
 *   
 *   Copyright (C) 2000-2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "mol_config.h"
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/poll.h>
#include <sys/resource.h>
#ifdef CONFIG_ALSA
#define ALSA_PCM_NEW_HW_PARAMS_API
//#define ALSA_PCM_NEW_SW_PARAMS_API
#include <alsa/asoundlib.h>
#endif
#include "driver_mgr.h"
#include "debugger.h"
#include "os_interface.h"
#include "verbose.h"
#include "memory.h"
#include "thread.h"
#include "res_manager.h"
#include "osi_driver.h"
#include "pic.h"
#include "async.h"
#include "booter.h"
#include "pci.h"
#include "timer.h"
#include "mac_registers.h"
#include "sound_sh.h"


typedef struct {
	char	*name;
	void	(*cleanup)( void );

	int	(*query_mode)( int format, int rate, int *frasize );
	int	(*open)( int format, int rate, int *fragsize );
	void	(*close)( void );

	void	(*write)( char *fragptr, int size );
	void	(*write_nb)( char *fragptr, int size );
	void	(*flush)( void );
	int	(*get_pollfd)( void );
} sound_ops_t;


static struct {
#ifdef CONFIG_ALSA
	snd_pcm_t	*pcm;		/* ALSA */
#endif
	int		fd;		/* OSS */

	int		irq;
	int		async_id;

	int		started;

	/* misc */
	int		rate_limit;	/* user imposed rate restriction */

	/* current mode */
	int		mute;
	int		format;
	uint		rate;		/* in Hz */
	int		bpf;		/* bytes per frame */

	/* sound parameters */
	int		fragsize;	/* client fragsize */

	/* output driver */
	sound_ops_t	*ops;

	/* ring buffer */
	char		*ringbuf;	/* ringbuf (if non-NULL) */
	int		ringfrags;	/* #frags in ringbuf */

	timestamp_t	timestamp;

	volatile uint	ringind;	/* last frag passed to hardware */
	volatile int	stop_ring;
} ss;



/************************************************************************/
/*	/dev/dsp support						*/
/************************************************************************/

static int
oss_convert_format( int format, int *fmt, int *channels )
{
	static struct {
		int mol_fmt, linux_fmt;
	} fmt_table[] = {
		{ kSoundFormat_U8_BE,	AFMT_U8,	},
		{ kSoundFormat_S16_BE,	AFMT_S16_BE,	},
		{ kSoundFormat_S16_LE,	AFMT_S16_LE,	},
	};
	int f, i;	

	for( f=-1, i=0; i<sizeof(fmt_table)/sizeof(fmt_table[0]); i++ )
		if( format & fmt_table[i].mol_fmt )
			f = fmt_table[i].linux_fmt;
	if( f ==  -1 )
		return -1;

	*fmt = f;
	*channels = (format & kSoundFormat_Stereo)? 2:1;
	return 0;
}

static void
oss_close( void )
{
	close( ss.fd );
	ss.fd = -1;
}

static const char *
oss_devname( void ) 
{
	char *s;
	if( !(s=get_filename_res("sound_device")) )
		s = "/dev/dsp";
	return s;
}

static int
oss_open( int format, int rate, int *fragsize )
{
	int size, err, fmt, channels, rate_, fmt_, channels_;
	uint frag = (4 << 16) | 12;	/* 4 x 4K */

	if( is_osx_boot() ) {
		//frag = (4 << 16) | 9;	/* 4 x 512 */
		frag = (4 << 16) | 8;	/* 4 x 256 */
	}

	if( ss.started )
		return 1;

	if( ss.fd != -1 || oss_convert_format( format, &fmt, &channels ))
		return -1;

	if( (ss.fd=open(oss_devname(), O_WRONLY | O_NONBLOCK)) < 0 )
		return 1;

	err=0;
	fmt_ = fmt; channels_ = channels; rate_ = rate;
	if( ioctl( ss.fd, SNDCTL_DSP_SETFMT, &fmt ) < 0 || fmt_ != fmt )
		err--;

	if( ioctl( ss.fd, SNDCTL_DSP_CHANNELS, &channels ) < 0 || channels_ != channels )
		err--;

	if( ioctl(ss.fd, SNDCTL_DSP_SPEED, &rate) < 0 || rate_ != rate )
		err--;

	ioctl( ss.fd, SNDCTL_DSP_SETFRAGMENT, &frag );
	ioctl( ss.fd, SNDCTL_DSP_GETBLKSIZE, &size );

	if( (1 << (frag & 0xffff)) != size ) {
		printm("Warning: fragsize not accepted\n");
		size = 32 * 1024;
	}
	if( fragsize )
		*fragsize = size;

	if( err ) {
		close( ss.fd );
		ss.fd = -1;
	}
	return err;
}

/* 0: ok, 1: might be supported, <0: unsupported/error */
static int
oss_query_mode( int format, int rate, int *fragsize )
{
	int ret;
	if( (ret=oss_open(format, rate, fragsize)) )
		return ret;
	oss_close();
	return 0;
}

static void
oss_flush( void )
{
	if( ioctl(ss.fd, SNDCTL_DSP_POST) < 0 )
		perrorm("SNDCTL_DSP_POST\n");
}

/* size should _always_ equal ss.fragsize */
static void
oss_write( char *buf, int size )
{
	struct pollfd ufd;

	ufd.fd = ss.fd;
	ufd.events = POLLOUT;

	poll( &ufd, 1, -1 );
	write( ss.fd, buf, size );
}

static void
oss_write_nb( char *buf, int size )
{
	write( ss.fd, buf, size );
}

static int
oss_get_pollfd( void ) 
{
	return ss.fd;
}

static sound_ops_t oss_driver_ops = {
	.name		= "oss",
	.cleanup	= NULL,
	.query_mode	= oss_query_mode,
	.open		= oss_open,
	.close 		= oss_close,
	.write		= oss_write,
	.write_nb	= oss_write_nb,
	.flush		= oss_flush,
	.get_pollfd	= oss_get_pollfd,
};

static sound_ops_t *
oss_probe( void )
{
	/* FIXME */
	return &oss_driver_ops;
}


/************************************************************************/
/*	ALSA support							*/
/************************************************************************/

#ifdef CONFIG_ALSA
#define ALSA_FRAGSIZE	512

static int
alsa_query_mode( int format, int rate, int *fragsize )
{
	if( rate != 44100 && rate != 22050 )
		return -1;
	if( !(format & kSoundFormat_S16_BE) )
		return -1;
	if( fragsize )
		*fragsize = ALSA_FRAGSIZE;
	return 0;
}

static void
alsa_prefill( void )
{
	snd_pcm_status_t *status;
	char buf[512];
	int frames, cnt;

	snd_pcm_status_malloc( &status );
	snd_pcm_status( ss.pcm, status );
	frames = snd_pcm_status_get_avail(status);
	snd_pcm_status_free( status );

	printm("Frames: %d\n", frames );
	
	memset( buf, 0, sizeof(buf) );
	cnt = sizeof(buf)/ss.bpf;
	for( ; frames > 0; frames -= cnt  ) {
		if( cnt > frames )
			cnt = frames;
		snd_pcm_writei( ss.pcm, buf, sizeof(buf)/ss.bpf );
	}
}

static int
alsa_open( int format, int rate, int *fragsize )
{
	snd_pcm_hw_params_t *params;

	printm("ALSA rate %d\n", rate );
	if( snd_pcm_open(&ss.pcm, "default", SND_PCM_STREAM_PLAYBACK, 0 * SND_PCM_NONBLOCK ) < 0 )
		return -1;

	/* XXX: fix cleanup */
	if( snd_pcm_hw_params_malloc(&params) < 0 )
		return -1;
	snd_pcm_hw_params_any( ss.pcm, params );
	snd_pcm_hw_params_set_access( ss.pcm, params, SND_PCM_ACCESS_RW_INTERLEAVED );
	snd_pcm_hw_params_set_format( ss.pcm, params, SND_PCM_FORMAT_S16_BE );
	snd_pcm_hw_params_set_rate_near( ss.pcm, params, &rate, 0 );
	snd_pcm_hw_params_set_channels( ss.pcm, params, 2 );
	snd_pcm_hw_params( ss.pcm, params );

	snd_pcm_hw_params_free( params );
	snd_pcm_prepare( ss.pcm );

	alsa_prefill();
	
	if( fragsize )
		*fragsize = ALSA_FRAGSIZE;

	if( snd_pcm_poll_descriptors_count(ss.pcm) != 1 )
		printm("ALSA: more than 1 poll descriptor\n");

	/* increase priority... */
	setpriority( PRIO_PROCESS, getpid(), -12 );
	return 0;
}

static void
alsa_close( void )
{
	printm("ALSA-close\n");
	snd_pcm_close( ss.pcm );
	ss.pcm = NULL;

	/* restore priority */
	setpriority( PRIO_PROCESS, getpid(), 0 );
}

static void
alsa_write( char *buf, int size ) 
{
	struct pollfd pfd;

	snd_pcm_poll_descriptors( ss.pcm, &pfd, 1 );
	poll( &pfd, 1, -1 );

	snd_pcm_writei( ss.pcm, buf, size/4 );
}

static void
alsa_write_nb( char *buf, int size ) 
{
	struct pollfd pfd;

	snd_pcm_poll_descriptors( ss.pcm, &pfd, 1 );
	poll( &pfd, 1, -1 );

	snd_pcm_writei( ss.pcm, buf, size/4 );
}

static void
alsa_flush( void ) 
{
}

static int
alsa_get_pollfd( void )
{
	struct pollfd pfd;

	snd_pcm_poll_descriptors( ss.pcm, &pfd, 1 );
	return pfd.fd;
}

static sound_ops_t alsa_driver_ops = {
	.name		= "osp",
	.cleanup	= NULL,
	.query_mode	= alsa_query_mode,
	.open		= alsa_open,
	.close 		= alsa_close,
	.write		= alsa_write,
	.write_nb	= alsa_write_nb,
	.flush		= alsa_flush,
	.get_pollfd	= alsa_get_pollfd,
};

static sound_ops_t *
alsa_probe( void )
{
	snd_pcm_t *pcm;

	if( snd_pcm_open(&pcm, "default", SND_PCM_STREAM_PLAYBACK, 1 * SND_PCM_NONBLOCK ) < 0 ) {
		printm("ALSA sound driver unavailable\n");
		return NULL;
	}
	snd_pcm_close( pcm );
	printm("ALSA sound driver\n");
	return &alsa_driver_ops;
}
#else
static sound_ops_t *
alsa_probe( void )
{
	return NULL;
}
#endif /* CONFIG_ALSA */

/************************************************************************/
/*	generic routines						*/
/************************************************************************/

static int
query_mode( int format, int rate )
{
	if( ss.rate_limit && rate > ss.rate_limit )
		return -1;
	return (*ss.ops->query_mode)( format, rate, NULL );
}

static int
set_mode( int format, int rate )
{
	int ret, fragsize;
	
	if( ss.started )
		return 1;
	if( (ret=(*ss.ops->query_mode)(format, rate, &fragsize)) )
		return ret;
	/* printm("SoundMode: %x, %d Hz, fragsize %d\n", format, rate, fragsize ); */
	
	ss.rate = rate;
	ss.bpf = (format & kSoundFormat_Len1Mask) ? 1 : 2;
	ss.bpf *= (format & kSoundFormat_Stereo ) ? 2 : 1;
	ss.fragsize = fragsize;
	ss.format = format;
	return 0;
}

static void
sndpipe_empty( int fd, int events ) 
{
	/* disable async handler until IRQ is raised... */
	set_async_handler_events( ss.async_id, 0 );
	irq_line_hi( ss.irq );
}

static void
ringbuf_thread( void *dummy )
{
	char *p;
	
	setpriority( PRIO_PROCESS, getpid(), -20 );
	while( !ss.stop_ring ) {
		p = ss.ringbuf + ss.ringind * ss.fragsize;

		(*ss.ops->write)( p, ss.fragsize );

		if( ++ss.ringind == ss.ringfrags ) {
			ss.ringind = 0;
			get_timestamp( &ss.timestamp );
			irq_line_hi( ss.irq );
		} else {
			abort_doze();
		}
	}
	setpriority( PRIO_PROCESS, getpid(), 0 );
	ss.stop_ring = -1;
}

static void
stop_sound( void )
{
	if( !ss.started )
		return;

	if( ss.ringbuf ) {
		ss.stop_ring = 1;

		while( ss.stop_ring != -1 )
			;
	} else {
		delete_async_handler( ss.async_id );
	}
	(*ss.ops->close)();
	ss.started = 0;
}

static int
start_sound( void ) 
{
	if( ss.started || !ss.format )
		return -1;

	if( (*ss.ops->open)(ss.format, ss.rate, NULL) ) {
		/* use timer fallback? */
		return -1;
	}
	ss.started = 1;

	if( ss.ringbuf ) {
		ss.ringind = 0;
		ss.stop_ring = 0;

		create_thread( (void*)ringbuf_thread, NULL, "audio-thread" );
	} else {
		/* XXX side-effect: sets O_NONBLOCK */
		int fd = (*ss.ops->get_pollfd)();
		ss.async_id = add_async_handler( fd, 0, sndpipe_empty, 0 );
	}
	return 0;
}

static void
flush_sound( void )
{
	if( !ss.started )
		return;
	(*ss.ops->flush)();
}

static void
resume_sound( void )
{
	if( ss.started )
		set_async_handler_events( ss.async_id, POLLOUT );
}

static void
pause_sound( void )
{
	if( ss.started )
		set_async_handler_events( ss.async_id, 0 );
}

static int
query_pipefull( void )
{
	/* unimplemented */
	return -1;
}



/************************************************************************/
/*	OSI interface							*/
/************************************************************************/

static int
osip_sound_irq_ack( int sel, int *params )
{
	mregs->gpr[4] = ss.timestamp.hi;
	mregs->gpr[5] = ss.timestamp.lo;
	return irq_line_low( ss.irq );
}

static int
osip_sound_cntl( int sel, int *params )
{
	switch( params[0] ) {
	/* ---- Version Checking (new API) ---- */
	case kSoundCheckAPI:
		if( params[1] != OSI_SOUND_API_VERSION ) {
			printm("Incompatible MOLAudio version!\n");
			return -1;
		}
		break;
	case kSoundOSXDriverVersion:
		set_driver_ok_flag( kOSXAudioDriverOK, params[1] == OSX_SOUND_DRIVER_VERSION );
		break;

	/* ---- New API (OSX) ---- */
	case kSoundSetupRing: /* bugfize, buf */
		ss.ringfrags = params[1] / ss.fragsize;
		ss.ringbuf = params[2] ? transl_mphys(params[2]) : NULL;
		break;
	case kSoundGetRingSample: 
		return ss.ringind * ss.fragsize/ss.bpf;

	case kSoundGetSampleOffset:
		/* no-touch zone, in frames */
		return 0;
	case kSoundGetSampleLatency:
		/* internal buffering, in frames */
		return 2048;

	/* ---- Both new and old API ---- */
	case kSoundQueryFormat: /* format, rate */
		return query_mode( params[1], params[2] );
	case kSoundSetFormat: /* format, rate */
		return set_mode( params[1], params[2] );
	case kSoundStart:
		return start_sound();
	case kSoundStop:
		stop_sound();
		break;
	case kSoundFlush:
		flush_sound();
		break;

	/* ---- Old API (classic) ---- */
	case kSoundClassicDriverVersion:
		if( params[1] < CLASSIC_SOUND_DRIVER_VERSION )
			printm("Install the latest MOLAudio extension!\n");
		break;
	case kSoundResume:
		resume_sound();
		break;
	case kSoundPause:
		pause_sound();
		break;
	case kSoundGetFragsize:
		return ss.fragsize;		
	case kSoundRouteIRQ:
		oldworld_route_irq( params[1], &ss.irq, "sound" );
		break;
	case kSoundQueryPipeFull: /* pipe full? */
		return query_pipefull();
	default:
		return -1;
	}
	return 0;
}

/* phys_buffer, len, restart */
static int
osip_sound_write( int sel, int *params )
{
	char *buf = transl_mphys( params[0] );
	int len = params[1];
	
	if( !ss.started || !buf )
		return -1;

	/* pad... buffer is guaranteed to be at least of size ss.fragsize */
	while( len < ss.fragsize )
		buf[len++] = 0;

	(*ss.ops->write_nb)( buf, ss.fragsize );

	if( params[2] )
		set_async_handler_events( ss.async_id, POLLOUT );
	return 0;
}

/* hwVolume, speakerVolume, hwMute */
static int
osip_volume( int sel, int *params )
{
	uint vol, lvol = (params[0] & 0xffff0000) >> 16;
	uint rvol = params[0] & 0xffff;
	ss.mute = params[2] ? 1:0;

	if( ss.fd < 0 )
		return 0;

	if( (lvol=lvol*100/0x100) > 100 )
		lvol = 100;
	if( (rvol=rvol*100/0x100) > 100 )
		rvol = 100;
	vol = rvol*0x100 | lvol;

	/* printm("Volume %d/%d\n", lvol, rvol ); */

	ioctl( ss.fd, SOUND_MIXER_WRITE_VOLUME, &vol );
	return 0;
}

#if 0
static void
startboing( void *dummy )
{
	char *name = get_filename_res("startboing_file");
	int format = AFMT_S16_BE;
	int rate = 22050;
	int channels = 2;

	if( (fd=open(name, O_RDONLY)) < 0 ) {
		printm("Could not open '%s'", name);
		return;
	}
}
#endif

/************************************************************************/
/*	init / cleanup							*/
/************************************************************************/

static osi_iface_t osi_iface[] = {
	{ OSI_SOUND_CNTL,	osip_sound_cntl		},
	{ OSI_SOUND_IRQ_ACK,	osip_sound_irq_ack	},
	{ OSI_SOUND_WRITE,	osip_sound_write	},
	{ OSI_SOUND_SET_VOLUME,	osip_volume 		},
};

static int 
osi_sound_init( void )
{
	static pci_dev_info_t pci_config = { 0x6666, 0x6668, 0x02, 0x0, 0x0000 };
	char *drv = is_oldworld_boot() ? "sound" : NULL;
	int i, limit;
	
	memset( &ss, 0, sizeof(ss) );
	ss.fd = -1;

	if( !register_osi_driver( drv, "mol-audio", is_classic_boot()? &pci_config : NULL, &ss.irq ) ) {
		printm("Failed to register sound card\n");
		return 0;
	}

	/* handle user rate limits */
	if( (limit=get_numeric_res("max_sound_rate")) > 0 )
		ss.rate_limit = limit;

	/* select output driver */
	for( i=0 ;; i++ ) {
		char *s = get_str_res_ind("sound_driver", 0, i );
		
		if( i && !s )
			break;
		if( s && !strcasecmp("any", s) )
			s = NULL;
		if( (!s || !strcasecmp("alsa", s)) && (ss.ops=alsa_probe()) )
			break;
		if( (!s || !strcasecmp("oss", s)) && (ss.ops=oss_probe()) )
			break;
	}
	/* XXX: fixme */
	if( !ss.ops )
		ss.ops = &oss_driver_ops;

	/* the osi procs must be registered even if sound is unavailable */
	register_osi_iface( osi_iface, sizeof(osi_iface) );
	return 1;
}

static void
osi_sound_cleanup( void )
{
}

driver_interface_t osi_sound_driver = {
    "osi_sound", osi_sound_init, osi_sound_cleanup
};
